﻿Imports System
'using System.Threading.Tasks; // optional for TPL parallelization

' library of static matrix methods including decomposition, inverse, determinant

Namespace MatrixDecomposition
	Friend Class MatrixDecompositionProgram
		Shared Sub Main(ByVal args() As String)
			Console.WriteLine(vbLf & "Begin matrix decomposition demo" & vbLf)

			Dim m()() As Double = MatrixCreate(4, 4)
			m(0)(0) = 3.0
			m(0)(1) = 7.0
			m(0)(2) = 2.0
			m(0)(3) = 5.0
			m(1)(0) = 1.0
			m(1)(1) = 8.0
			m(1)(2) = 4.0
			m(1)(3) = 2.0
			m(2)(0) = 2.0
			m(2)(1) = 1.0
			m(2)(2) = 9.0
			m(2)(3) = 3.0
			m(3)(0) = 5.0
			m(3)(1) = 4.0
			m(3)(2) = 7.0
			m(3)(3) = 1.0

			Console.WriteLine("Matrix m = " & vbLf & MatrixAsString(m))

			Dim perm() As Integer = New Integer() {}
			Dim toggle As Integer
			Dim luMatrix()() As Double = MatrixDecompose(m, perm, toggle)
			Dim lower()() As Double = ExtractLower(luMatrix)
			Dim upper()() As Double = ExtractUpper(luMatrix)

			Console.WriteLine("The (combined) LUP decomposition of m is" & vbLf & MatrixAsString(luMatrix))
			Console.WriteLine("The decomposition permutation array is: " & VectorAsString(perm))
			Console.WriteLine("The decomposition toggle value is: " & toggle)
			Console.WriteLine(vbLf & "The lower part of LU is " & vbLf & MatrixAsString(lower))
			Console.WriteLine("The upper part of LU is " & vbLf & MatrixAsString(upper))

			Dim inverse()() As Double = MatrixInverse(m)
			Console.WriteLine("Inverse of m computed from its decomposition is" & vbLf & MatrixAsString(inverse))

			Dim det As Double = MatrixDeterminant(m)
			Console.WriteLine("Determinant of m computed via decomposition = " & det.ToString("F1"))

			Console.WriteLine(vbLf & "Linear system mx = b problem: " & vbLf)
			Console.WriteLine("3x0 + 7x1 + 2x2 + 5x3 = 49")
			Console.WriteLine(" x0 + 8x1 + 4x2 + 2x3 = 30")
			Console.WriteLine("2x0 +  x1 + 9x2 + 3x3 = 43")
			Console.WriteLine("5x0 + 4x1 + 7x2 +  x3 = 52")
			Console.WriteLine("")

			Dim b() As Double = {49.0, 30.0, 43.0, 52.0}
			Console.WriteLine("Solving system via decomposition")
			Dim x() As Double = SystemSolve(m, b)
			Console.WriteLine(vbLf & "Solution is x = " & vbLf & VectorAsString(x))

			' extra: matrix decomposition concepts
			Dim lu()() As Double = MatrixProduct(lower, upper)
			Dim orig()() As Double = UnPermute(lu, perm) ' use a custom method with the perm array to unscramble LU
			If MatrixAreEqual(orig, m, 0.000000001) = True Then
				Console.WriteLine("Product of L and U successfully unpermuted using perm array")
			End If

			Dim permMatrix()() As Double = PermArrayToMatrix(perm) ' convert the perm array to a matrix
			orig = MatrixProduct(permMatrix, lu) ' another way to unscramble
			If MatrixAreEqual(orig, m, 0.000000001) = True Then
				Console.WriteLine(vbLf & "Product of L and U successfully unpermuted using perm matrix" & vbLf)
			End If

			Console.WriteLine("End matrix decomposition demo" & vbLf)
			Console.ReadLine()
		End Sub

		' static matrix methods - consider placing in a class.

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixCreate(ByVal rows As Integer, ByVal cols As Integer) As Double()()
			' allocates/creates a matrix initialized to all 0.0. assume rows and cols > 0
			' do error checking here
			Dim result(rows - 1)() As Double
			For i As Integer = 0 To rows - 1
				result(i) = New Double(cols - 1) {}
			Next i

			'For i As Integer = 0 To rows - 1
			'	For j As Integer = 0 To cols - 1
			'		result(i)(j) = 0.0 ' explicit initialization needed in some languages
			'	Next j
			'Next i

			Return result
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixRandom(ByVal rows As Integer, ByVal cols As Integer, ByVal minVal As Double, ByVal maxVal As Double, ByVal seed As Integer) As Double()()
			' return a matrix with random values
			Dim ran As New Random(seed)
			Dim result()() As Double = MatrixCreate(rows, cols)
			For i As Integer = 0 To rows - 1
				For j As Integer = 0 To cols - 1
					result(i)(j) = (maxVal - minVal) * ran.NextDouble() + minVal
				Next j
			Next i
			Return result
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixIdentity(ByVal n As Integer) As Double()()
			' return an n x n Identity matrix
			Dim result()() As Double = MatrixCreate(n, n)
			For i As Integer = 0 To n - 1
				result(i)(i) = 1.0
			Next i

			Return result
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixAsString(ByVal matrix()() As Double) As String
			Dim s As String = ""
			For i As Integer = 0 To matrix.Length - 1
				For j As Integer = 0 To matrix(i).Length - 1
					s &= matrix(i)(j).ToString("F3").PadLeft(8) & " "
				Next j
				s &= Environment.NewLine
			Next i
			Return s
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixAreEqual(ByVal matrixA()() As Double, ByVal matrixB()() As Double, ByVal epsilon As Double) As Boolean
			' true if all values in matrixA == corresponding values in matrixB
			Dim aRows As Integer = matrixA.Length
			Dim aCols As Integer = matrixA(0).Length
			Dim bRows As Integer = matrixB.Length
			Dim bCols As Integer = matrixB(0).Length
			If aRows <> bRows OrElse aCols <> bCols Then
				Throw New Exception("Non-conformable matrices in MatrixAreEqual")
			End If

			For i As Integer = 0 To aRows - 1 ' each row of A and B
				For j As Integer = 0 To aCols - 1 ' each col of A and B
					'If matrixA(i)(j) <> matrixB(i)(j) Then
					If Math.Abs(matrixA(i)(j) - matrixB(i)(j)) > epsilon Then
						Return False
					End If
				Next j
			Next i
			Return True
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixProduct(ByVal matrixA()() As Double, ByVal matrixB()() As Double) As Double()()
			Dim aRows As Integer = matrixA.Length
			Dim aCols As Integer = matrixA(0).Length
			Dim bRows As Integer = matrixB.Length
			Dim bCols As Integer = matrixB(0).Length
			If aCols <> bRows Then
				Throw New Exception("Non-conformable matrices in MatrixProduct")
			End If

			Dim result()() As Double = MatrixCreate(aRows, bCols)

			For i As Integer = 0 To aRows - 1 ' each row of A
				For j As Integer = 0 To bCols - 1 ' each col of B
					For k As Integer = 0 To aCols - 1 ' could use k < bRows
						result(i)(j) += matrixA(i)(k) * matrixB(k)(j)
					Next k
				Next j
			Next i

			'Parallel.For(0, aRows, Sub(i)
			'						   For j As Integer = 0 To bCols - 1 ' each col of B
			'							   For k As Integer = 0 To aCols - 1 ' could use k < bRows
			'								   result(i)(j) += matrixA(i)(k) * matrixB(k)(j)
			'							   Next k
			'						   Next j
			'					   End Sub);

			Return result
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixVectorProduct(ByVal matrix()() As Double, ByVal vector() As Double) As Double()
			' result of multiplying an n x m matrix by a m x 1 column vector (yielding an n x 1 column vector)
			Dim mRows As Integer = matrix.Length
			Dim mCols As Integer = matrix(0).Length
			Dim vRows As Integer = vector.Length
			If mCols <> vRows Then
				Throw New Exception("Non-conformable matrix and vector in MatrixVectorProduct")
			End If
			Dim result(mRows - 1) As Double	' an n x m matrix times a m x 1 column vector is a n x 1 column vector
			For i As Integer = 0 To mRows - 1
				For j As Integer = 0 To mCols - 1
					result(i) += matrix(i)(j) * vector(j)
				Next j
			Next i
			Return result
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixDecompose(ByVal matrix()() As Double, <System.Runtime.InteropServices.Out()> ByRef perm() As Integer, <System.Runtime.InteropServices.Out()> ByRef toggle As Integer) As Double()()
			' Doolittle LUP decomposition with partial pivoting.
			' rerturns: result is L (with 1s on diagonal) and U; perm holds row permutations; toggle is +1 or -1 (even or odd)
			Dim rows As Integer = matrix.Length
			Dim cols As Integer = matrix(0).Length ' assume all rows have the same number of columns so just use row [0].
			If rows <> cols Then
				Throw New Exception("Attempt to MatrixDecompose a non-square mattrix")
			End If

			Dim n As Integer = rows	' convenience

			Dim result()() As Double = MatrixDuplicate(matrix) ' make a copy of the input matrix

			perm = New Integer(n - 1) {} ' set up row permutation result
			For i As Integer = 0 To n - 1
				perm(i) = i
			Next i

			toggle = 1 ' toggle tracks row swaps. +1 -> even, -1 -> odd. used by MatrixDeterminant

			For j As Integer = 0 To n - 2 ' each column
				Dim colMax As Double = Math.Abs(result(j)(j)) ' find largest value in col j
				Dim pRow As Integer = j
				For i As Integer = j + 1 To n - 1
					If result(i)(j) > colMax Then
						colMax = result(i)(j)
						pRow = i
					End If
				Next i

				If pRow <> j Then ' if largest value not on pivot, swap rows
					Dim rowPtr() As Double = result(pRow)
					result(pRow) = result(j)
					result(j) = rowPtr

					Dim tmp As Integer = perm(pRow)	' and swap perm info
					perm(pRow) = perm(j)
					perm(j) = tmp

					toggle = -toggle ' adjust the row-swap toggle
				End If

				If Math.Abs(result(j)(j)) < 1.0E-20 Then ' if diagonal after swap is zero...
					Return Nothing ' consider a throw
				End If

				For i As Integer = j + 1 To n - 1
					result(i)(j) /= result(j)(j)
					For k As Integer = j + 1 To n - 1
						result(i)(k) -= result(i)(j) * result(j)(k)
					Next k
				Next i
			Next j ' main j column loop

			Return result
		End Function ' MatrixDecompose

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixInverse(ByVal matrix()() As Double) As Double()()
			Dim n As Integer = matrix.Length
			Dim result()() As Double = MatrixDuplicate(matrix)

			Dim perm() As Integer = New Integer() {}
			Dim toggle As Integer
			Dim lum()() As Double = MatrixDecompose(matrix, perm, toggle)
			If lum Is Nothing Then
				Throw New Exception("Unable to compute inverse")
			End If

			Dim b(n - 1) As Double
			For i As Integer = 0 To n - 1
				For j As Integer = 0 To n - 1
					If i = perm(j) Then
						b(j) = 1.0
					Else
						b(j) = 0.0
					End If
				Next j

				Dim x() As Double = HelperSolve(lum, b)

				For j As Integer = 0 To n - 1
					result(j)(i) = x(j)
				Next j
			Next i
			Return result
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixDeterminant(ByVal matrix()() As Double) As Double
			Dim perm() As Integer = New Integer() {}
			Dim toggle As Integer
			Dim lum()() As Double = MatrixDecompose(matrix, perm, toggle)
			If lum Is Nothing Then
				Throw New Exception("Unable to compute MatrixDeterminant")
			End If
			Dim result As Double = toggle
			For i As Integer = 0 To lum.Length - 1
				result *= lum(i)(i)
			Next i
			Return result
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function HelperSolve(ByVal luMatrix()() As Double, ByVal b() As Double) As Double() ' helper
			' before calling this helper, permute b using the perm array from MatrixDecompose that generated luMatrix
			Dim n As Integer = luMatrix.Length
			Dim x(n - 1) As Double
			b.CopyTo(x, 0)

			For i As Integer = 1 To n - 1
				Dim sum As Double = x(i)
				For j As Integer = 0 To i - 1
					sum -= luMatrix(i)(j) * x(j)
				Next j
				x(i) = sum
			Next i

			x(n - 1) /= luMatrix(n - 1)(n - 1)
			For i As Integer = n - 2 To 0 Step -1
				Dim sum As Double = x(i)
				For j As Integer = i + 1 To n - 1
					sum -= luMatrix(i)(j) * x(j)
				Next j
				x(i) = sum / luMatrix(i)(i)
			Next i

			Return x
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function SystemSolve(ByVal A()() As Double, ByVal b() As Double) As Double()
			' Solve Ax = b
			Dim n As Integer = A.Length

			' 1. decompose A
			Dim perm() As Integer = New Integer() {}
			Dim toggle As Integer
			Dim luMatrix()() As Double = MatrixDecompose(A, perm, toggle)
			If luMatrix Is Nothing Then
				Return Nothing
			End If

			' 2. permute b according to perm[] into bp
			Dim bp(b.Length - 1) As Double
			For i As Integer = 0 To n - 1
				bp(i) = b(perm(i))
			Next i

			' 3. call helper
			Dim x() As Double = HelperSolve(luMatrix, bp)
			Return x
		End Function ' SystemSolve

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function MatrixDuplicate(ByVal matrix()() As Double) As Double()()
			' allocates/creates a duplicate of a matrix. assumes matrix is not null.
			Dim result()() As Double = MatrixCreate(matrix.Length, matrix(0).Length)
			For i As Integer = 0 To matrix.Length - 1 ' copy the values
				For j As Integer = 0 To matrix(i).Length - 1
					result(i)(j) = matrix(i)(j)
				Next j
			Next i
			Return result
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function ExtractLower(ByVal matrix()() As Double) As Double()()
			' lower part of a Doolittle decomposition (1.0s on diagonal, 0.0s in upper)
			Dim rows As Integer = matrix.Length
			Dim cols As Integer = matrix(0).Length
			Dim result()() As Double = MatrixCreate(rows, cols)
			For i As Integer = 0 To rows - 1
				For j As Integer = 0 To cols - 1
					If i = j Then
						result(i)(j) = 1.0
					ElseIf i > j Then
						result(i)(j) = matrix(i)(j)
					End If
				Next j
			Next i
			Return result
		End Function

		Private Shared Function ExtractUpper(ByVal matrix()() As Double) As Double()()
			' upper part of a Doolittle decomposition (0.0s in the strictly lower part)
			Dim rows As Integer = matrix.Length
			Dim cols As Integer = matrix(0).Length
			Dim result()() As Double = MatrixCreate(rows, cols)
			For i As Integer = 0 To rows - 1
				For j As Integer = 0 To cols - 1
					If i <= j Then
						result(i)(j) = matrix(i)(j)
					End If
				Next j
			Next i
			Return result
		End Function

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function PermArrayToMatrix(ByVal perm() As Integer) As Double()()
			' convert Doolittle perm array to corresponding perm matrix
			Dim n As Integer = perm.Length
			Dim result()() As Double = MatrixCreate(n, n)
			For i As Integer = 0 To n - 1
				result(i)(perm(i)) = 1.0
			Next i
			Return result
		End Function

		Private Shared Function UnPermute(ByVal luProduct()() As Double, ByVal perm() As Integer) As Double()()
			' unpermute product of Doolittle lower * upper matrix according to perm[]
			' no real use except to demo LU decomposition, or for consistency testing
			Dim result()() As Double = MatrixDuplicate(luProduct)

			Dim unperm(perm.Length - 1) As Integer
			For i As Integer = 0 To perm.Length - 1
				unperm(perm(i)) = i
			Next i

			For r As Integer = 0 To luProduct.Length - 1
				result(r) = luProduct(unperm(r))
			Next r

			Return result
		End Function ' UnPermute

		' --------------------------------------------------------------------------------------------------------------

		Private Shared Function VectorAsString(ByVal vector() As Double) As String
			Dim s As String = ""
			For i As Integer = 0 To vector.Length - 1
				s &= vector(i).ToString("F3").PadLeft(8) & Environment.NewLine
			Next i
			s &= Environment.NewLine
			Return s
		End Function

		Private Shared Function VectorAsString(ByVal vector() As Integer) As String
			Dim s As String = ""
			For i As Integer = 0 To vector.Length - 1
				s &= vector(i).ToString().PadLeft(2) & " "
			Next i
			s &= Environment.NewLine
			Return s
		End Function

		' --------------------------------------------------------------------------------------------------------------

	End Class ' class MatrixDecompositionProgram

End Namespace ' ns